/*
	elementals.js (c) Jason M. Knight
	Version 3.2 PRODUCTION, 14 November 2017
*/

/*
	*** MAINTAINER NOTES ***

	do NOT use "this" inside user callable objects like _.Class or _.Node as
	"this" is invalid during certain types of callbacks, such as Array.each
*/

// *** START primary elementals "underscore" object ***
var _ = {

	BROWSER_PREFIXES : ['ms', 'moz', 'o', 'webkit'],
	
	Class : {

		add : function(element, cName) {
			return cName instanceof Array ? cName.callbackOr(element, _.Class.add) : (
				_.Class.exists(element, cName) ||
				(element.className += (element.className ? ' ' : '') + cName)
			);
		}, // _.Class.add

		exists : function(element, cName) {
			return cName instanceof Array ? cName.callbackAnd(element, _.Class.exists) : (
				RegExp('(\\s|^)' + cName + '(\\s|$)').test(element.className)
			);
		}, // _.Class.exists

		remove : function(element, cName) {
			if (cName instanceof Array) return cName.callbackOr(element, _.Class.remove);
			if (!_.Class.exists(element, cName)) return false;
			element.className = element.className.replace(
				new RegExp('(\\s|^)' + cName + '(\\s|$)'), ' '
			).trim();
			return true;
		}, // _.Class.remove

		swap : function(element, oldClass, newClass) {
			oldClass = _.Class.remove(element, oldClass);
			newClass = _.Class.add(element, newClass);
			return oldClass || newClass;
		}, // _.Class.swap

		toggle : function(element, cName) {
			if (_.Class.remove(element, cName)) return false;
			_.Class.add(element, cName);
			return true;
		} // _.Class.toggle

	}, // _.Class

	Compare : {

		string : function(a, b) {
			return _.normalize(a).localeCompare(
				_.normalize(b)
			);
		},

		stringI : function(a, b) {
			return _.normalize(a, true).localeCompare(
				_.normalize(b, true)
			);
		},

		numeric : function(a, b) {
			return (Number(a) || 0) - (Number(b) || 0);
		}

	}, // _.Compare

	dataGet : function(element, name, splitResult, temp) {
		return (temp = element.getAttribute(name = 'data-' + name)) === null ? false :
			temp === '' ? element.hasAttribute && element.hasAttribute(name) :
				splitResult ? temp.split(',') : temp;
	}, // _.dataGet

	EMPTY_TAGS : [
		'AREA', 'BASE', 'BR', 'COL', 'COMMAND', 'EMBED', 'HR', 'IMG',
		'INPUT', 'KEYGEN', 'LINK', 'META', 'PARAM', 'SOURCE', 'TRACK',
		'WBR'
	], // _.EMPTY_TAGS

	normalize : function(stringish, toLowerCase) {
		stringish = (
			'undefined' == typeof stringish ||
			(!stringish && stringish !== 0)
		) ? '' : String(stringish).trim();
		return toLowerCase ? stringish.tolowercase() : stringish;
	}, // _.normalize

	Throw : {

		isNaN : function(value) {
			_.Throw.isNull(value);
			if (value != value) throw new RangeError('Value must be a number');
		}, // _.Throw.isNaN

		isNull : function(value) {
			if (null == value || 'undefined' == typeof value) throw new TypeError(
				'Variable Null or Undefined'
			);
		}, // _.Throw.isNull

		nonElement : function(element) {
			_.Throw.isNull(element);
			if (element.nodeType !== 1) throw new TypeError(
				element + ' is not a DOM Element'
			);
		}, // _.Throw.nonElement

		nonInstance : function(object1, object2) {
			_.Throw.isNull(object1);
			if (!(object1 instanceof object2)) throw new TypeError(
				object1 + ' is not an instance of ' + object2
			);
		}, // _.Throw.nonInstance

		nonType : function(value, type) {
			_.Throw.isNull(value);
			if (type !== typeof value) throw new TypeError(
				value + ' is not a "' + type + '" type'
			);
		}, // _.Throw.nonType

		notTagName : function(element, tagName) {
			_.Throw.isNull(element);
			if (
				1 == !element.nodeType ||
				element.tagName.toUpperCase() != tagName.toUpperCase()
			) throw new TypeError(
				element + ' is not a <' + tagName + '>'
			);
		}, // _.Throw.notTagName

		range : function(value, min, max) {
			_.Throw.isNaN(value);
			if (value == Infinity) throw new RangeError(
				'Value must be less than Inifinty'
			);
			if (!isNaN(min) && value < min) throw new RangeError(
				'Value must be greater than or equal to ' + min
			);
			if (!isNaN(max) && value > max) throw new RangeError(
				'Value must be less than or equal to ' + max
			);
		} // _.Throw.range

	} // _.Throw

}; // _
// *** END primary elementals "underscore" object ***

// *** START self-instancing function for local subfunctoons and variables ***

(function() {

	"use strict";

	// *** START local variables ***
	var
		w = window,
		d = document,
		db = d.body,
		de = d.documentElement,
		ua = navigator.userAgent,
		ua_msie = ua.match(/MSIE ([0-9\.]+)/i),
		ua_edge = ua.match(/Edge\/([0-9\.]+)/i),
		ua_rv = ua.match(/rv:([0-9\.]+)/i),
		escMap = {
			'"' : '\\"',
			'\\' : '\\\\',
			'\b' : '\\b',
			'\f' : '\\f',
			'\n' : '\\n',
			'\r' : '\\r',
			'\t' : '\\t'
		},
		escRegex = /[\\\u0000-\u001F\u0022\u2028\u2029]/g,
		htmlSpecialCharsMap = {
			'&' : '&amp;',
			'<' : '&lt;',
			'>' : '&gt;',
			'"' : '&quot;',
			"'" : '&#039;'
		},
		htmlSpecialCharsRegex = /[&<>"']/g,
		stripCamelRegex = /([A-Z])/g,
		toCamelRegex = /(\W[\w])/g,
		trimRegex = /^[\s\uFEFF\xA0]+|[\s\uFEFF\xA0]+$/g,
		tempList,
		EventAdd,
	// *** END local variables ***

	// *** START local objects ***
	
		writes = {
			first : function(element, content) {
				return element.insertBefore(content, element.firstChild);
			},
			after : function(element, content) {
				return element.parentNode.insertBefore(content, element.nextSibling);
			},
			before : function(element, content) {
				return element.parentNode.insertBefore(content, element);
			},
			replace : function(element, content) {
				return _.Node.replace(element, content);
			},
			rewrite : function(element, content) {
				_.Node.flush(element);
				return writes.last(element, content);
			},
			last : function(element, content) {
				return element.appendChild(content);
			}
		};
	
	// *** END local objects

	/*
		*** START local functions ***

		These functions are invisible outside of this self-instancing
		function and are meant for internal use by elementals ONLY.
	*/

	function browserFix(target, methodName, altMethodName) {
		if (target[methodName]) return true;
		altMethodName || (altMethodName = methodName.charAt(0).toUpperCase() + methodName.substr(1));
		for (var i = 0, j, k; j = _.BROWSER_PREFIXES[i]; i++)
			if (target[k = j + altMethodName]) return target[methodName] = target[k], true;
		return false;
	} // browserFix

	function eXMLHttp(params, ajax) {

		if (!(ajax = this.result = w.XMLHttpRequest ? new XMLHttpRequest() : (
			newAX('Msxml2.XMLHTTP.6.0') ||
			newAX('Msxml2.XMLHTTP.3.0') ||
			newAX('Microsoft.XMLHTTP')
		))) throw new Error('Unable to create XMLHttpRequest or ActiveX XMLHTTP');

		if (params.progress) result.onprogress = result.progress;

		ajax.getJSON = function() {
			return JSON.parse(this.responseText);
		};

		ajax.getXMLValueById = function(id) {
			return (id = this.responseXML.getElementById(id)) ? _.Node.text(id) : null;
		};

		ajax.getXMLValuesByTagName = function(tagname, limit, elements) {
			if (!(elements = this.responseXML.getElementsByTagName(tagName))) return false;
			if (true == limit) return _.Node.text(elements[0]);
			if (!limit || elements.length < limit) limit = elements.length;
			for (var result = [], i = 0; i < limit; i++) result.push(_.Node.text(elements[i]));
			return result;
		};

		ajax.onreadystatechange  = function() {
			if (
				params.state &&
				params.state[ajax.readyState]
			) return params.state[ajax.readyState](ajax, params);
			if (4 == ajax.readyState) {
				if (params.status) {
					if (params.status[ajax.status]) return params.status[ajax.status](ajax, params);
					if (params.status.catchAll) return params.status.catchAll(ajax, params);
				}
				_.Log[params.onStatusLog || 'warn']('Unhandled AJAX Status ' + ajax.status);
			} else if (params.onReadyLog) _.Log[params.onReadyLog]('Unhandled AJAX readyState ' + ajax.readyState);
		};

		if (params.request) {
			ajax.open(params.request[0], params.request[1]);
			if (params.mimeType) ajax.overrideMimeType(params.mimeType);
			if (params.headers)
				for (var i in params.headers)
					ajax.setRequestHeader(i, params.headers[i]);
			ajax.send(params.request[2]);
		}

	} // eXMLHttp

	function makeSelectionButton(targetElement, buttonContent, buttonParent, placement, callback, temp) {
		_.Throw.nonElement(targetElement);
		if (_.NON_CONTENT_TAGS.indexOf(targetElement.tagName.toUpperCase()) !== -1)
			_Log.Error('Cannot create selection button for <' + targetElement.tagName + '>');
		temp = _.make('button.selectContents', { content : buttonContent, onclick : callback });
		temp['data-eNodeSelectTarget'] = targetElement;
		_.Node.write(buttonParent || targetElement, temp, placement || 'before');
	} // makeSelectionButton

	function newAX(axName) {
		try { return new activeXObject(axName); } catch (axName) { return false; }
	} // newAX

	function nCountBy(element, nodeMask, method, i) {
		i = 0;
		_.Node[method](element, function() { return i++ }, nodeMask);
		return i;
	} // nCountBy

	function originalSort(a, b) {
		return _.Compare.numeric(
			a.tag.getAttribute('data-USortOriginalPlace'),
			b.tag.getAttribute('data-USortOriginalPlace')
		);
	} // originalSort

	function polyfillAddPrototype(targetObject, sources) {
		for (var i in sources)
			targetObject.prototype[i] || (targetObject.prototype[i] = sources[i]);
	} // polyfillAdd
	
	function polyfillAddDirect(targetObject, sources) {
		for (var i in sources)
			targetObject[i] || (targetObject[i] = sources[i]);
	} // polyfillAddDirect

	function USortTR(tr, result) {
		tr.getAttribute('data-USortOriginalPlace') ||
		tr.setAttribute('data-USortOriginalPlace', tempList.length);
		result = [];
		_.Node.eachChild(tr, function(a) {
			return result.push(_.Node.text(a)), true;
		});
		tempList.push({ tag : tr.cloneNode(true), data : result });
		return true;
	} // USortTR

	function USortLI(li) {
		li.getAttribute('data-USortOriginalPlace') ||
		li.setAttribute('data-USortOriginalPlace', tempList.length);
		tempList.push({ tag : li.cloneNode(true), data : _.Node.text(li) });
		return true;
	} // USortLI
	
	function writeExistsIn(obj) {
		for (var i in obj) if (writes[i]) return true;
		return false;
	} // writeExistsIn
	
	// *** END local functions ***
	
	// *** START underscore properties that require locals ***
	
	_.IS_IE = ua_msie ? ua_msie[1] : (/Trident/i.test(ua) ? ua_rv[1] : false);
	_.IS_EDGE = (ua_edge && !_.IS_IE) ? ua_edge[1] : false;
	
	// *** END underscore properties that require locals ***
	
	// *** START underscore methods that require locals ***

	_.ajax = function(params) {
		switch (typeof params) {
			case 'function':
				params = { 'status' : { '200' : params } };
			case 'object':
				return (new eXMLHttp(params)).result;
		}
		throw new Error(
			'Invalid Ajax declaration, parameter "handler" not a function or object!'
		);
	}; // _.ajax

	_.Cookies = {

		/*
			technically 'delete' is a reserved word, so we have to make it
			a string to get IE8/earlier to use it as a method name. DERP
		*/

		'delete' : function(name, path, domain) {
			_.Cookies.get(name) && _.Cookies.set(name, '', 0, path, domain);
		}, // _.Cookies.delete

		get : function(name) {
			return name && (name = d.cookie.match(
				new RegExp('(^|;)\\s*' + name + '\\s*=\\s*([^;]*)')
			) ? unescape(name[2]) : null);
		}, // _.Cookies.get

		set : function(name, value, time, path, domain, secure) {
			if ('undefined' == typeof time) time = 34160760000000;
			if (Number.isInteger(time)) time = new Date(time);
			if (time instanceof Date) time = time.toGMTString();
			_.Throw.nonType(time, 'string');
			d.cookie = name + '=' + escape(value) +
				'; expires=' + time +
				(path ? '; path=' + path : '') +
				(domain ? '; domain=' + domain : '') +
				(secure ? '; secure' : '');
		}, // _.Cookies.set

		setAfter : function(name, value, time, path, domain, secure) {
			_.Cookies.set(name, value, time ? new Date().getTime() + time : false, path, domain, secure);
		} // _.Cookies.setAfter

	}; // _.Cookies

	_.Event = {

		add : function(e, value, callback, prevent) {
			new EventAdd(e, value, callback, true == prevent ? 1 : (prevent || 0));
		},

		PREVENT_BUBBLE : 1,
		PREVENT_PROPAGATION : 3

	}; // _.Event

	// START browser specific _.Event.prevent and local EventAdd object
	if (w.addEventListener) {
		_.Event.prevent = function(e) {
			if (e.stopPropagation) e.stopPropagation();
			if (e.preventDefault) e.preventDefault();
		};
		EventAdd = function(element, event, callback, captureLevel) {
			element.addEventListener(event, function(e) {
				callback(e);
				if (captureLevel > 1) _.Event.prevent(e);
			}, captureLevel == 1);
		};
	} else {
		_.Event.prevent = function(e) {
			e.cancelBubble = true;
			e.returnValue = false;
		};
		EventAdd = function(element, event, callback, captureLevel) {
			element.attachEvent('on' + event, function(e) {
				e = w.event;
				e.target = e.srcElement;
				e.currentTarget = element;
				callback(e);
				if (captureLevel > 0) {
					e.cancelBubble = true;
					if (captureLevel > 1) e.returnValue = false;
				}
			});
		};
	}
	// END browser specific _.Event.prevent and local EventAdd object

	_.formSerialize = function(formElement) {

		_.Throw.notTagName(formElement, 'FORM');

		function add(name, value) {
			if (value && (value != '')) result +=
				(result ? '&' : '') +
				encodeURIComponent(name) +
				'=' +
				encodeURIComponent(value);
		}

		for (
			var i = 0, result = '', temp, e;
			e = formElement.elements[i];
			i++
		) {
			if (e.name) switch (e.tagName.toUpperCase()) {
				case 'INPUT':
					switch (e.type) {
						case 'file' :
							// to-do, add serialization for browsers that support it
							throw new TypeError('Cannot Serialize a <input type="file">');
						case 'checkbox':
						case 'radio':
							if (!e.checked) continue;
					}
				case 'BUTTON':
				case 'TEXTAREA':
					add(e.name, e.value);
				break;
				case 'SELECT':
					temp = e.options[e.selectedIndex];
					add(e.name, temp.value || _.Node.text(temp));
			}
		}

		return result;

	}; // _.formSerialize

	_.Load = {

		style : function(href, media) {
			return media ? ( media = media.toLowerCase(), "all" == media && _.Log.warn(
				'media="all" is NOT recommended - ' + href
			)) : _.Log.warn(
				'no media type declared for ' + href + ', defaulting to "screen,projection,tv"'
			), _.make('link', {
				rel : 'stylesheet',
				href : href,
				media : media || 'screen,projection,tv',
				last : d.head
			}
		)}, // _.Load.style

		script : function(src, attr) {
			return _.make('script@' + src, attr);
		} // _.Load.script

	}; // _.Load

	_.Log = {

		alert : function(message) {
			alert(message);
		}, // _.Log.alert

		append : function(message) {
			_.make('br', { last : db });
			_.Node.write(db, message);
		}, // _.Log.append

		error : function(message) {
			throw new Error(a);
		}, // _.Log.error

		log : function(message) {
			w.console ? console.log(message) : _.Log.append(message);
		}, // _.Log.log

		warn : function(message) {
			w.console ? console.warn(message) : _.Log.append(a);
		} // _.Log.warn

	}; // _.Log

	_.make = function(selector, attributes, e) {
		selector = selector.splitSelector();
		if (selector.tag == 'TEXTNODE') e = d.createTextNode(selector.special || '');
		else {
			e = d.createElement(selector.tag);
			selector.classes && (e.className = selector.classes);
			selector.id && (e.id = selector.id);
			if (selector.special) switch (selector.tag.toLowerCase()) {
				case 'a': e.href = selector.special; break;
				case 'button':
				case 'input':
					e.type = selector.special;
					break;
				case 'script':
					e.src = selector.special;
					if (!attributes) attributes = { last : db };
					else if (!writeExistsIn(attributes)) attributes.last = db;
					break;
				case 'th': e.scope = selector.special; break;
				case 'td': e.headers = selector.special; break;
			}
		}
		if ("object" == typeof attributes) {
			for (var name in attributes) {
				var value = attributes[name];
				if ('on' == name.substr(0, 2)) _.Event.add(e, name.substr(2), value);
				if (writes[name]) _.Node.write(value, e, name);
				else switch (name) {
					case 'children':
						for (var i = 0, child; child = value[i]; i++) {
							if (
								"undefined" != typeof child[1] &&
								"object" != typeof child[1]
							) child[1] = { content : child[1] };
							if (!child[1]) child[1] = { last : e };
							else if (!writeExistsIn(child[1])) child[1].last = e;
							_.make(child[0], child[1]);
						}
						break;
					case 'content':
						_.Node.write(e, value, name);
						break;
					case 'style':
						for (var i in value) e.style[i] = value[i];
						break;
					case 'type':
						// legacy IE can't handle setting type on Element directly
						e.setAttribute(name, value);
						break;
					default:
						// but legacy IE screws up setAttribute a LOT on other attributes so...
						e[name] = value;
				}
			}
		} else if (attributes) _.Node.write(e, attributes);
		return e;
	}; // _.make

	_.Metrics = {

		absoluteX : function(element) {
			return element.getBoundingClientRect().left + _.Metrics.Scroll.x();
		}, // _.Metrics.absoluteX

		absoluteY : function(element) {
			return element.getBoundingClientRect().top + _.Metrics.Scroll.y();
		}, // _.Metrics.absoluteY

		height : function(element) {
			return !element || element == w ? (
				w.innerHeight || de.clientHeight || db.clientHeight
			) : element.clientHeight;
		}, // _.Metrics.height

		Scroll : {

			height : function(element) {
				return (!element || element == w ? db : element).scrollHeight;
			}, // _.Metrics.Scroll.height

			maxX : function(element) {
				return _.Metrics.Scroll.width(element) - _.Metrics.width(element);
			}, // _.Metrics.Scroll.maxX

			maxY : function(element) {
				return _.Metrics.Scroll.height(element) - _.Metrics.height(element);
			}, // _.Metrics.Scroll.maxY

			width : function(element) {
				return (!element || element == w ? db : element).scrollWidth;
			}, // _.Metrics.Scroll.width

			x : function(element) {
				return !element || element == w ? (
					w.scrollX || w.pageXOffset || db.scrollLeft
				) : element.scrollLeft;
			}, // _.Metrics.Scroll.x

			y : function(element) {
				return !element || element == w ? (
					w.scrollY || w.pageYOffset || db.scrollTop
				) : e.scrollTop;
			} // _.Metrics.Scroll.y

		}, // _.Metrics.Scroll

		width : function(element) {
			return !element || element == w ? (
				w.innerWidth || de.clientWidth || db.clientWidth
			) : element.clientWidth;
		} // _.Metrics.width

	}; // _.Metrics

	_.Node = {

		addCopyButton : function(
			element, buttonContent, buttonParentElement, placement
		) {
			makeSelectionButton(
				element,
				buttonContent || 'Copy',
				buttonParentElement,
				placement,
				function(a) {
					_.Select.node(a.currentTarget['data-eNodeSelectTarget']);
					_.Select.toClipboard();
				}
			);
		}, // _.Node.addCopyButton

		addSelectButton : function(
			element, buttonContent, buttonParentElement, placement
		) {
			makeSelectionButton(
				element,
				buttonContent || 'Select',
				buttonParentElement,
				placement,
				function(e) {
					_.Select.node(e.currentTarget['data-eNodeSelectTarget']);
				}
			);
		}, // _.Node.addSelectButton

		countAll : function(element, nodeMask) {
			return nCountBy(element, nodeMask, 'eachAll');
		}, // _.Node.countAll

		countChildren : function(element, nodeMask) {
			return nCountBy(element, nodeMask, 'eachChild');
		}, // _.Node.countChildren

		eachAll : function(element, callback, nodeMask, temp) {
			if (temp = element.firstChild) do {
				if (
					(_.Node.mask(temp, nodeMask) && false == callback(temp)) ||
					(temp.firstChild && false == _.Node.eachAll(temp, callback, nodeMask))
				) return false;
			} while (temp = temp.nextSibling);
			return true;
		}, // _.Node.eachAll

		eachChild : function(element, callback, nodeMask, temp) {
			if (temp = element.firstChild) do {
				if (
					_.Node.mask(temp, nodeMask) &&
					false == callback(temp)
				) return false;
			} while (temp = temp.nextSibling);
			return true;
		}, // _.Node.eachChild

		first : function(element, nodeMask) {
			return _.Node.walk(element, nodeMask, 'firstChild', 'nextSibling');
		}, // _.Node.first

		flush : function(element) {
			while (element.firstChild) element.removeChild(element.firstChild);
		}, // _.Node.flush

		last : function(element, nodeMask) {
			return _.Node.walk(element, nodeMask, 'lastChild', 'previousSibling');
		}, // _.Node.last

		mask : function(element, nodeMask) {
			return isNaN(nodeMask = nodeMask || 1) ? (
				element.nodeType == 1 &&
				element.tagName.toUpperCase() == nodeMask.toUpperCase()
			) : ((
				element.nodeType == 0 ? 0 : 1 << (element.nodeType - 1)
			) & nodeMask) > 0;
		}, // _.Node.mask

		next : function(element, nodeMask) {
			return _.Node.walk(element, nodeMask, 'nextSibling', 'nextSibling');
		}, // _.Node.next

		nextRemove : function(element, nodeMask, temp) {
			temp = __.Node.next(element, nodeMask);
			_.Node.remove(element);
			return temp;
		}, // _.Node.nextRemove

		prev : function(element, nodeMask) {
			return _.Node.walk(element, nodeMask, 'previousSibling', 'previousSibling');
		}, // _.Node.prev

		prevRemove : function(element, nodeMask, temp) {
			temp = _.Node.prev(element, nodeMask);
			_.Node.remove(element);
			return temp;
		}, // _.Node.prevRemove

		remove : function(element) {
			element.parentNode.removeChild(element);
		}, // _.Node.remove

		replace : function(oldNode, newNode) {
			return oldNode.parentNode.replaceChild(
				'object' == typeof newNode ? newNode : d.createTextNode(newNode),
				oldNode
			);
		}, // _.Node.replace

		text : function(element) {
			return element ? element.innerText || element.textContent : false;
		}, // _.Node.text

		walk : function(element, nodeMask, firstMethod, nextMethod, temp) {
			if (temp = element[firstMethod]) do {
				if (_.Node.mask(temp, nodeMask)) return temp;
			} while (temp = temp[nextMethod]);
			return false;
		}, // _.Node.walk
		
		write : function(element, content, placement) {
			writes[writes[placement] ? placement : 'last'](
				element,
				'object' == typeof content ? content : d.createTextNode(content)
			);
		} // _.Node.write
		
	}; // _.Node

	_.Node.mask.ELEMENT    = 1;
	_.Node.mask.TEXT       = 4;
	_.Node.mask.PROCESSING = 64;
	_.Node.mask.COMMENT    = 128;
	_.Node.mask.DOCUMENT   = 256;
	_.Node.mask.DOCTYPE    = 512;
	_.Node.mask.FRAGMENT   = 1024;
	
	_.NON_CONTENT_TAGS = _.EMPTY_TAGS.concat([
		'AUDIO', 'COLGROUP', 'HEAD', 'MAP', 'METER', 'PROGRESS',
		'OBJECT', 'SCRIPT', 'STYLE', 'TEXTAREA', 'VIDEO'
	]); // _.NON_CONTENT_TAGS

	_.Select = {

		all : function() {
			_.Select.node(db);
		} // _.Select.all

	}; // _.Select

	// START browser specific _.Select.node and _.Select.none
	if (d.selection) {
		_.Select.node = function(element, temp) {
			temp = db.createTextRange();
			temp.moveToElementText(element);
			temp.select();
			return true;
		};
		_.Select.none = function() {
			d.selection.empty();
			return true;
		};
	} else if (w.getSelection) {
		_.Select.node = function(element, tempRange, tempSel) {
			tempSel = w.getSelection();
			tempRange = d.createRange();
			tempRange.selectNodeContents(element);
			tempSel.removeAllRanges();
			tempSel.addRange(tempRange);
			return true;
		};
		_.Select.none = function() {
			w.getSelection().removeAllRanges();
			return true;
		};
	} else {
		_.Select.node =
		_.Node.writeSelectButton = function() {
			_.Log.warn('selection of content');
			return false;
		};
		_.Select.none = function() {
			_.Log.warn('de-selection of content');
			return false;
		};
	}
	// END browser specific _.Select.node and _.Select.none

	// START browser specific _.writeToClipboard, _.Select.toClipBoard
	if (
		w.clipboardData &&
		w.clipboardData.setData
	) {
		_.writeToClipboard = function(a) {
			w.clipboardData.setData('Text', String(a));
			return true;
		};
		_.Select.toClipboard = function() {
			_.writeToClipboard(
				'function' == typeof d.getSelection ?
				String(d.getSelection()) :
				d.selection.createRange().text
			);
			return true;
		};
	} else if (d.queryCommandSupported('copy')) {
		_.writeToClipboard = function(a, b) {
			b = make('textarea', { content : String(a), last : db });
			b.select();
			d.execCommand('copy');
			_.Node.remove(b);
			return true;
		};
		_.Select.toClipboard = function() {
			d.execCommand('copy');
			return true;
		};
	} else {
		_.writeToClipboard =
		_.Select.toClipboard = function() {
			_.Log.warn('This browser does not allow JavaScript to access the clipboard');
			return false;
		};
	}
	// END browser specific _.writeToClipboard, _.Select.toClipBoard

	_.USort = {

		element : function(e, callback) {
			tempList = [];
			switch (e.tagName.toUpperCase()) {
				case 'TBODY' :
					_.Node.eachChild(e, USortTR, 'TR');
					break;
				case 'UL':
				case 'OL':
					_.Node.eachChild(e, USortLI, 'LI');
					break;
				default: throw new TypeError(e + ' is not a TBODY, UL, or OL');
			}
			e.setAttribute('data-USortElementSorted', true);
			e['data-USortElementSorted'] = true;
			tempList.sort(callback);
			_.Node.flush(e);
			for (var i = 0, j; j = tempList[i]; i++) e.appendChild(j.tag);
			tempList = null;
		}, // _.USort.element

		restore : function(e) {
			switch (e.tagName.toUpperCase()) {
				case 'TBODY':
				case 'UL':
				case 'OL':
					if (e['data-USortElementSorted']) _.USort.element(e, originalSort);
					return;
			}
			throw new TypeError(e + ' is not a TBODY, UL, or OL');
		} // _.USort.restore

	}; // _.USort

	// START browser specific _.Query and _.Query.all
	if ('function' == typeof d.querySelectorAll) {
		_.Query = function(element, selector) {
			return element.querySelector(selector);
		};
		_.Query.all = function(element, selector) {
			// [].slice makes it return an ARRAY
			return [].slice.call(element.querySelectorAll(selector));
		};
	} else {
		var tempStyle = d.createStyleSheet();
		_.Query = function(element, selector, temp) {
			temp = _.Query.all(element, selector, 1);
			return (temp && temp.length > 0) ? temp[0] : null;
		};
		_.Query.all = function(element, selector, limit, results) {
			if (!limit) limit = 32767;
			results = [];
			tempStyle.addRule(selector, 'fu:bar');
			function callback(e) {
				if ((limit > 0) && (e.currentStyle.fu === 'bar')) results.push(e);
				return --limit > 0;
			}
			_.Node.eachAll(element, callback);
			tempStyle.removeRule();
			return results;
		};
	}
	// END browser specific _.Query and _.Query.all

	// *** END underscore methods that require locals ***

	// *** START polyfills and browser object extensions ***

	polyfillAddPrototype(Array, {

		callbackAnd : function(a, b) {
			for (var i = 0, l = this.length; i < l; i++)
				if (!b(a, this[i])) return false;
			return true;
		}, // Array.callbackAnd

		callbackAndAll : function(a, b) {
			for (var i = 0, l = this.length, r = false; i < l; i++)
				r = b(a, this[i]) && r;
			return r;
		}, // Array.callbackAndAll

		callbackOr : function(a, b) {
			for (var i = 0, l = this.length, r = false ; i < l; i++)
				r = b(a, this[i]) || r;
			return r;
		}, // Array.callbackOr

		every : function(a, b) {
			_.Throw.nonType(a, 'function');
			b = 'undefined' == typeof b ? void 0 : b;
			for (var i = 0, j = this.length; i < j; i++)
				if (!a.call(b, this[i], i, this)) return false;
			return true;
		}, // Array.every

		fill : function(a, b, c) {
			_.Throw.isNull(this);
			if ('undefined' == typeof b) b = 0;
				else if (b < 0) b += this.length;
			if ('undefined' == typeof c) c = this.length;
			for (var i = b; i < c; i++) this[i] = a;
			return this;
		}, // Array.fill

		filter : function(a, b) {
			_.Throw.nonType(a, 'function');
			b = 'undefined' == typeof b ? void 0 : b;
			for (var r = [], i = 0, j = this.length; i < j; i++)
				if (a.call(b, this[i], i, this)) r.push(this[i]);
			return r;
		}, // Array.filter

		find : function(a, b) {
			_.Throw.nonType(a, 'function');
			b = 'undefined' == typeof b ? void 0 : b;
			for (var i = 0, j = this.length; i < j; i++)
				if (a.call(b, this[i], i, this)) return this[i];
		}, // Array.find

		findIndex : function(a, b) {
			_.Throw.nonType(a, 'function');
			b = 'undefined' == typeof b ? void 0 : b;
			for (var i = 0, j = this.length; i < j; i++)
				if (a.call(b, this[i], i, this)) return i;
		}, // Array.findIndex

		forEach : function(a, b) {
			_.Throw.nonType(a, 'function');
			b = 'undefined' == typeof b ? void 0 : b;
			for (var i = 0, j = this.length; i < j; i++)
				a.call(b, this[i], i, this);
		}, // Array.forEach

		includes : function(a, b) {
			_.Throw.isNull(this);
			if (this.length == 0) return false;
			if ('undefined' == typeof b || (b < 0 && (b += this.length) < 0)) b = 0;
			var c = a != a;
			for (var j = this.length; b < j; b++)
				if (a == this[b] || (c && this[b] != this[b])) return true;
			return false;
		}, // Array.includes

		indexOf : function(a, b) {
			_.Throw.isNull(this);
			if (this.length == 0) return false;
			if ('undefined' == typeof b || (b < 0 && (b += this.length) < 0)) b = 0;
			for (var c = this.length; b < c; b++) if (this[b] == a) return b;
			return -1;
		}, // Array.indexOf

		isArray : function(a) {
			return Object.prototype.toString.call(a) === '[object Array]';
		}, // Array.isArray

		lastIndexOf : function(a, b) {
			_.Throw.isNull(this);
			if (this.length == 0) return false;
			if (
				'undefined' == typeof b || (b < 0 && (b += this.length) < 0)
			) b = this.length;
			while (b > 0) if (this[--b] == a) return b;
			return -1;
		}, // Array.lastIndexOf

		map : function(a, b) {
			_.Throw.nonType(a, 'function');
			b = 'undefined' == typeof b ? void 0 : b;
			for (var r = [], i = 0, j = this.length; i < j; i++)
				r.push(a.call(b, this[i], i, this));
			return r;
		}, // Array.map

		reduce : function(a, b) {
			_.Throw.nonType(a, 'function');
			var i = 0, j = this.length;
			if ('undefined' == typeof b) {
				if (!j) throw new TypeError(
					'Reduce of empty array with no initial value'
				);
				if (j == 1) return this[0];
				b = this[i];
				i++;
			}
			while (i < j) b = a(v, this[i], i++, this);
			return b;
		}, // Array.reduce

		reduceRight : function(a, b) {
			_.Throw.nonType(a, 'function');
			var i = this.length;
			if ('undefined' == typeof b) {
				if (!i) throw new TypeError(
					'Reduce of empty array with no initial value'
				);
				if (i == 1) return this[0];
				b = this[--i];
			}
			while (i > 0) b = a(b, this[--i], i, this);
			return b;
		}, // Array.reduceRight

		some : function(a, b) {
			_.Throw.nonType(a, 'function');
			b = 'undefined' == typeof b ? void 0 : b;
			for (var i = 0, j = this.length; i < j; i++)
				if (a.call(b, this[i], i, this)) return true;
			return false;
		} // Array.some

	}); // polyfill Array
	
	// 'of' has to go on it directly.
	polyfillAddDirect(Array, {
		'of' : function() {
			_.Throw.isNull(this);
			return Array.prototype.slice.call(arguments);
		}
	});
	
	polyfillAddDirect(Math, {
	
		// 'constants' 
		RAD : Math.PI / 180,
		QUD : Math.PI / 360,
		DEG : 180 / Math.PI,
		
		// methods
		
		acosh : function(a) {
			return Math.log(a + Math.sqrt(a * a - 1));
		}, // Math.acosh
		
		asinh : function(a) {
			return x === -Infinity ? x : Math.log(x + Math.sqrt(x * x + 1));
		}, // Math.asinh
		
		atanh : function(a) {
			return Math.log((1 + a) / (1 - a)) / 2;
		}, // Math.atanh
		
		cbrt : function(a, b) {
			b = Math.pow(Math.abs(a), 1 / 3);
			return a < 0 ? -b : b;
		}, // Math.cbrt
		
		clz32 : function(a) {
			return !a ? 32 : 31 - Math.floor(Math.log(x >>> 0) * Math.LOG2E);
		},
		
		clz64 : function(a) {
			return !a ? 64 : 63 - Math.floor(Math.log(x >>> 0) * Math.LOG2E);
		},
		
		cosh : function(a) {
			return ((a = Math.exp(a)) + 1 / a) / 2;
		},
		
		/*
			I am NOT polyfilling Math.expm1, type the -1 you lazy ****'s!
			Same for Math.log10, Math.logP, Math.log2
			Why the **** do these even EXIST?!?
			
			Skipping Math.fround, MAYBE later.
		*/
		
		hypot : function() {
			for (var i = 0, j = arguments.length, k, r = 0; i < j; i++) {
				if (
					Infinity == (k = arguments[i]) ||
					Infinity == -k
				) return Infinity;
				r += k * k;
			}
			return Math.sqrt(y);
		}, // Math.hypot
		
		imul : function(a, b) {
			var
				al = 0xFFFF & a,
				dl = 0xFFFF & b;
				return al * dl + ((( 
					(0xFFFF & (a >> 16)) * dl +
					(0xFFFF & (b >> 16)) * al
				) << 16) >>> 0) | 0;
		}, // Meth.imul
		
		sign : function(a) {
			return ((x > 0) - (x < 0)) || +x;
		}, // Math.sign
		
		sinh : function(a) {
			return ((a = Math.exp(a)) - 1 / a) / 2;
		},
		
		tanh : function(a, b, c) {
			return (b = Math.exp(+a)) == Infinity ? 1 : (
				(c = Math.exp(-a)) == Infinity ? -1 : (b - c) / (b + c)
			);
		},
		
		trunc : function (a) {
			return a < 0 ? Math.ceil(a) : Math.floor(a);
		}
		
	}); // polyfill Math
	
	polyfillAddDirect(Number, {
	
		EPSILON : 2.220446049250313e-16,
		MAX_SAFE_INTEGER : 9007199254740991,
		MIN_SAFE_INTEGER : -9007199254740991,
		
		isNaN : function(value) {
			return value != value;
		},
		
		'isFinite' : function(value) {
			return typeof value === 'number' && isFinite(value);
		},
		
		isSafeInteger : function(n) {
			return Number.isInteger(value) && Math.abs(value) <= Number.MAX_SAFE_INTEGER;
		}
		
	}); // polyfill Number
	
	polyfillAddPrototype(String, {
	
		htmlSpecialChars : function() {
			return text.replace(htmlSpecialCharsRegex, function(m) {
				return htmlSpecialCharsMap[m];
			});
		},
	
		endsWith : function(a, b) {
			_.Throw.isNull(this);
			if (b) _.Throw.isNaN(b);
			b = (
				(!b && b !== 0) ||
				b >= this.length
			) ? b.length : b;
			return this.substr(b - a.length, a.length) == a;
		}, // String.endsWith


		jsToCssName : function() {
			_.Throw.isNull(this);
			var
				v = this.stripCamel(),
				f = this.indexOf('-');
			return ((
				(f > 0) && (f = v.substr(0,f)) &&
				(_.BROWSER_PREFIXES.indexOf(f) !== -1)
			) ? '-' : '') + this;
		}, // String.jsToCssName
		
		pad : function(a, b, c) {
			_.Throw.isNaN(a);
			if ((c = Math.abs(a) - this.length) <= 0) return this;
			b = (!b && b !== 0) ? ' ' : String(b).charAt(0);
			return a < 0 ? b.repeat(c) + this : this + b.repeat(c);
		}, // String.pad
		
		padEnd : function(a, b) {
			return a <= 0 ? this : this.pad(a, b);
		},
		
		padStart : function(a, b) {
			return a <= 0 ? this : this.pad(-a, b);
		},

		repeat : function(a) {
			_.Throw.isNull(this);
			_.Throw.range(a = a || 0, 0, Math.floor(0x10000000 / this.length));
			return Array(Math.floor(a) + 1).join(this);
		}, // String.repeat

		splitSelector : function() {
			_.Throw.isNull(this);
			var
				r = { tag : 'div', id : false, classes : false, special : false},
				s = this, t;
			if ((t = s.indexOf('@')) >= 0) {
				r.special = s.substr(t + 1).replace('@', ' ');
				s = (t > 0) ? s.substr(0, t) : '';
			}
			if ((t = s.indexOf('.')) >= 0) {
				r.classes = s.substr(t + 1).replace('.', ' ');
				s = (t > 0) ? s.substr(0, t) : '';
			}
			if ((t = s.indexOf('#')) >= 0) {
				r.id = s.substr(t + 1);
				s = (t > 0) ? s.substr(0, t) : '';
			}
			if (s.length > 0) r.tag = s;
			return r;
		}, // String.splitSelector
		
		startsWith : function(a, b) {
			_.Throw.isNull(this);
			if (b) _.Throw.isNaN(b);
			return (a = String(a)) == this.substr(b || 0, a.length);
		}, // String.startsWith

		strFrom : function(a, b) {
			return (b = b  ? this.lastIndexOf(a) : this.indexOf(a)) == -1 ? false : this.substr(b + 1);
		}, // String.strFrom

		stripCamel : function(a) {
			if ('undefined' == typeof a) a = '-';
			return _.Throw.isNull(this), this.replace(stripCamelRegex, function(m) {
				return a + m.toLowerCase();
			});
		}, // String.stripCamel

		toCamel : function(a) {
			if ('undefined' == typeof a) a = '-';
			return _.Throw.isNull(this), this.replace(toCamelRegex, function(b, c, e) {
				return (e === 0 ? b : b.toUpperCase()).replace(a, '');
			});
		}, // String.toCamel

		trim : function() {
			return _.Throw.isNull(this), this.replace(trimRegex, '');
		} // String.trim

	}); // polyfill String

	// START JSON polyfill
	w.JSON || (w.JSON = {
		parse : function(value) { return eval('(' + value + ')'); },
		stringify : function(value, temp) {
			if (null == value) return null;
			switch (typeof value) {
				case 'number': return isFinite(value) ? value.toString() : 'null';
				case 'boolean': return value.toString();
				case 'object':
					if ('function' == typeof value.toJSON) return JSON.stringify(value.toJSON());
					temp = [];
					if (Array.prototype.isArray.call(value)) {
						value.forEach(function(c) { temp.push(JSON.stringify(c)); });
						return '[' + temp.join() + ']';
					}
					if (_.toString(value) === '[object Object]') {
						for (var i in value)
							temp.push(JSON.stringify(i) + ':' + JSON.stringify(value[i]));
						return '{' + temp.join() + '}';
					}
			}
			return '"' + value.toString().replace(escRegex, function(c) {
				return escMap[c] || '\\u' + (c.charCodeAt(0) + 65536).toString(16).substr(1);
			}) + '"';
		}
	});
	// END JSON Polyfill

	/*
		START performance and requestAnimation Polyfill
		Based on code by Erik Möller. I'd link but the bloody page is gone!
		Thanks "My Opera"
	*/
	var
		perfLastTime = 0,
		perf = false == 'performance' in w ? w.performance = {} : w.performance;

	if (false == 'now' in perf) {
		var nowOffset = (
			perf.timing && perf.timing.navigationStart
		) ? perf.timing.navigationStart : new Date().getTime();
		perf.now = function() { return new Date().getTime() - nowOffset; };
	}

	if (
		!browserFix(w, 'requestAnimationFrame')
	) w.requestAnimationFrame = function(callback, now, elapsed) {
		now = perf.now(),
		elapsed = Math.max(0, 16.666667 - now - perfLastTime);
		perfLastTime = now + elapsed;
		return setTimeout( function() { callback(perfLastTime); }, elapsed);
	};

	if (
		!browserFix(w, 'cancelAnimationFrame') &&
		!browserFix(w, 'cancelAnimationFrame', 'CancelRequestAnimationFrame')
	) w.cancelAnimationFrame = clearTimeout;
	// END performance and requestAnimation Polyfill

	// fix document.head if missing
	if (!d.head) d.head = d.getElementsByTagName('head')[0];

	// *** END polyfills and browser object extensions ***

	/*
		Add 'elementals' class to body to say "yes, elementals is running
		handy if we need it as a CSS or other such trigger.
	*/
	_.Class.add(db, 'elementals');

})();